use rand::distributions::StandardNormal;
use rand::prelude::ThreadRng;
use rand::{thread_rng, Rng};

use std::fmt::{Debug, Display, Formatter, Result};

use backtrack::{can_be_put, Reine};

pub struct AnnealingResult<T: Debug, U: Debug> {
    pub x: T,
    pub y: U,
    pub nb_iterations: usize,
    pub final_temp: f64,
}

impl<T: Debug, U: Debug> Display for AnnealingResult<T, U> {
    fn fmt(&self, f: &mut Formatter) -> Result {
        write!(
            f,
            "Solution found in {} iterations\n\nX={:?}\nY={:?}\nT={}",
            self.nb_iterations, self.x, self.y, self.final_temp
        )
    }
}

pub fn test_annealing() -> AnnealingResult<f64, f64> {
    const EPSILON: f64 = 1e-10;

    let f = |x: f64| {
        10.0 * ((0.3 * x) * (1.3 * x.powi(2) + 1e-5 * x.powi(4) + 0.2 * x + 80.0).sin()).sin()
    };

    let voisin = |x: f64| {
        let picked: f64 = thread_rng().sample(StandardNormal);

        x + picked
    };

    let decrease_temp = |t: f64| 0.99 * t;

    let accept = |delta: f64, t: f64| {
        if delta >= 0.0 {
            let a = (-delta / t).exp();

            return thread_rng().gen_range(0.0, 1.0) < a;
        }

        true
    };

    let mut x: f64 = 0.0;
    let mut t: f64 = 100.0;
    let n: usize = 100;

    let mut nb_iterations: usize = 0;

    while (f(x) + 10.0).abs() >= EPSILON {
        for _ in 0..n {
            let y: f64 = voisin(x);

            let delta: f64 = f(y) - f(x);

            if accept(delta, t) {
                x = y;
            }
        }

        nb_iterations += n;
        t = decrease_temp(t);
    }

    AnnealingResult {
        x: x,
        y: f(x),
        nb_iterations: nb_iterations,
        final_temp: t,
    }
}

pub fn nqueens_annealing(dim: usize, verbose: bool) -> AnnealingResult<Vec<Reine>, isize> {
    let mut rng = thread_rng();

    //numbers of blocked queens from being placed
    //(less is better, 0 is best)
    let f = |x: &Vec<Reine>| {
        let mut board: Vec<Reine> = Vec::new();

        let mut i = 0;

        for reine in x {
            if !can_be_put(*reine, &board) {
                i += 1;
            }

            board.push(*reine);
        }

        i
    };

    let voisin = |x: &Vec<Reine>, rng: &mut ThreadRng| {
        let mut to_swap: Vec<Reine> = x.clone();

        let first_index: usize = (rng.gen::<f64>() * x.len() as f64).abs() as usize;
        let mut second_index: usize = (rng.gen::<f64>() * x.len() as f64).abs() as usize;

        while second_index == first_index {
            second_index = (rng.gen::<f64>() * x.len() as f64).abs() as usize;
        }

        let tmp = to_swap[first_index].1;

        to_swap[first_index].1 = to_swap[second_index].1;
        to_swap[second_index].1 = tmp;

        to_swap
    };

    let decrease_temp = |t: f64| 0.991 * t;

    let accept = |delta: isize, t: f64| {
        if delta >= 0 {
            let a = (-delta as f64 / t).exp();

            return thread_rng().gen_range(0.0, 1.0) < a;
        }

        true
    };

    let mut x: Vec<Reine> = diagonal_chessboard(dim);
    let mut t: f64 = 10.0;
    let n: usize = 100;

    let mut nb_iterations: usize = 0;

    const BUF_LINE_COUNT: usize = 50;
    let mut buf_tracker: usize = 0;
    let mut print_buf: String = String::with_capacity(22 * BUF_LINE_COUNT);

    while f(&x) != 0 {
        for _ in 0..n {
            let y: Vec<Reine> = voisin(&x, &mut rng);

            let delta: isize = f(&y) - f(&x);

            if accept(delta, t) {
                x = y;
            }
        }

        nb_iterations += n;

        if verbose {
            buf_tracker += 1;

            print_buf += &format!("{:0>10} {:0>3.3} {:0>3}\n", nb_iterations, t, f(&x));

            if buf_tracker % BUF_LINE_COUNT == 0 {
                print!("{}", print_buf);
                print_buf.clear();
            }
        }

        t = decrease_temp(t);
    }

    let final_y = f(&x);

    AnnealingResult {
        x: x,
        y: final_y,
        nb_iterations: nb_iterations,
        final_temp: t,
    }
}

fn diagonal_chessboard(dim: usize) -> Vec<Reine> {
    let mut chessboard: Vec<Reine> = Vec::new();

    for i in 0..dim {
        chessboard.push((i as i32, i as i32));
    }

    chessboard
}
